/************************************************************************
 *   IRC - Internet Relay Chat, src/auth.c
 *   Copyright (C) 2003 Dmitry Agaphonov
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"

#include <dlfcn.h>

#define AUTHMODULE_NAME			"auth.so"
#define AUTHMODULE_INTERFACE_VERSION	0x0102

#define AMS_LOADED	0x0001

static const char denyreason[] = "Authorization failed, please contact your server administrator if you believe this is incorrect";

static void *am_handle = NULL;
static int am_status = 0;

static void (*am_on_unload_func)(void) = NULL;
static int (*am_on_registered_func)(const char *n, const char *u,
		const char *h, const char *g, const char *p,
		const struct in_addr *ia) = NULL;
static void (*am_version_func)(char **v) = NULL;
static void (*am_on_timer)(void) = NULL;

static int load_auth_module(char *errstr, size_t errlen)
{
    void (*am_ifver_func)(int *v) = NULL;
    int (*am_on_load_func)(void) = NULL;
    int (*am_get_timer)(void) = NULL;
    char buf[MAXPATHLEN];
    int ver;
    struct itimerval itv;

    if (am_status & AMS_LOADED)
    {
	snprintf(errstr, errlen, "Auth module is already loaded");
	return 0;
    }

    snprintf(buf, MAXPATHLEN, "%s/%s/%s", dpath, MDIR, AUTHMODULE_NAME);
#ifndef OS_OPENBSD
    if (!(am_handle = dlopen(buf, RTLD_NOW)))
#else
    if (!(am_handle = dlopen(buf, DL_LAZY)))
#endif
    {
	snprintf(errstr, errlen, "Error loading " AUTHMODULE_NAME ": %s",
			dlerror());
	return 0;
    }

    if (!(am_ifver_func = dlsym(am_handle, "get_interface_version")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    (*am_ifver_func)(&ver);
    if (ver != AUTHMODULE_INTERFACE_VERSION)
    {
	snprintf(errstr, errlen, AUTHMODULE_NAME " uses interface version "
			"%d, my interface version is %d", ver,
			AUTHMODULE_INTERFACE_VERSION);
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_on_load_func = dlsym(am_handle, "on_load")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_on_unload_func = dlsym(am_handle, "on_unload")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_version_func = dlsym(am_handle, "get_module_version")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_on_registered_func = dlsym(am_handle, "on_registered")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_get_timer = dlsym(am_handle, "get_timer")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(am_on_timer = dlsym(am_handle, "on_timer")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			AUTHMODULE_NAME ": %s", dlerror());
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    if (!(*am_on_load_func)())
    {
	snprintf(errstr, errlen, "Onload function in %s failed",
			AUTHMODULE_NAME);
	dlclose(am_handle);
	am_handle = NULL;
	return 0;
    }

    itv.it_interval.tv_sec = itv.it_value.tv_sec = (*am_get_timer)();
    itv.it_interval.tv_usec = itv.it_value.tv_usec = 0;
    setitimer(ITIMER_REAL, &itv, NULL);

    am_status |= AMS_LOADED;

    return 1;
}

static int unload_auth_module(char *errstr, size_t errlen)
{
    if (!(am_status & AMS_LOADED))
    {
	snprintf(errstr, errlen, "Auth module is not loaded");
	return 0;
    }

    (*am_on_unload_func)();

    dlclose(am_handle);
    am_handle = NULL;
    am_status &= ~AMS_LOADED;

    return 1;
}

void auth_init()
{
    char errbuf[256], *ver;

    if (!load_auth_module(errbuf, sizeof(errbuf)))
    {
	fprintf(stderr, "%s\n", errbuf);
	return;
    }

    (*am_version_func)(&ver);
    printf("Loaded module " AUTHMODULE_NAME ", version %s\n",
		    ver ? ver : "unspecified");
}

void auth_rehash(int sig)
{
    char errbuf[256], *ver;

    if (sig)
	sendto_ops("Got signal SIGUSR2, reloading auth module");

    if (!unload_auth_module(errbuf, sizeof(errbuf)))
	sendto_realops(errbuf);

    if (load_auth_module(errbuf, sizeof(errbuf)))
    {
	(*am_version_func)(&ver);
	sendto_realops("Successfully loaded module " AUTHMODULE_NAME ". "
			"Version: %s", ver ? ver : "unspecified");
    }
    else
	sendto_realops(errbuf);
}

int auth_check(aClient *cptr, const char *username)
{
    if (am_status & AMS_LOADED)
    {
	int rc = (*am_on_registered_func)(cptr->name, username,
		cptr->user->host, cptr->info, cptr->passwd, &cptr->ip);

	if (rc == -1)
	{
	    sendto_realops_lev(REJ_LEV, "Rejecting unauthorized client: %s "
			    "(%s@%s) [%s]", cptr->name, username,
			    cptr->user->host, cptr->info);

	    sendto_one(cptr, ":%s NOTICE %s :*** Notice -- %s", me.name,
		(BadPtr(cptr->name) ? "*" : cptr->name), denyreason);
	}

	return rc;
    }

    return 0;
}

char *auth_version(char *buf, int buflen)
{
    char *ver = NULL;

    if (!(am_status & AMS_LOADED))
    {
	ircsnprintf(buf, buflen, "No auth module is currently installed.");
	return buf;
    }

    (*am_version_func)(&ver);
    ircsnprintf(buf, buflen, "%s", ver ? ver : "unspecified");
    return buf;
}

void auth_timer()
{
    if (am_status & AMS_LOADED)
	(*am_on_timer)();
}
