/************************************************************************
 *   IRC - Internet Relay Chat, src/dnsbl.c
 *   Copyright (C) 2003 Dmitry Agaphonov
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: dnsbl.c 1270 2006-07-14 14:11:33Z andrew $ */

#include "struct.h"
#include "config.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "userban.h"
#include "throttle.h"
#include "dnsbl.h"
#include <arpa/inet.h>


aDNSBL	       *dnsbls = NULL;
Link	       *l_lines = NULL;

static Link    *ipfilters = NULL;
static Link    *hostfilters = NULL;

static int	hbl_count = 0;
static int	hf_count = 0;


static const char denyreason[] = "Your host is blacklisted";


static inline void update_blacklisted(aClient *cptr)
{
	if (!cptr->ipbls && (hbl_count || hf_count ?
		(cptr->blflags & BLFLAGS_GOTHOST ?
			cptr->blflags & BLFLAGS_SENTRHSS &&
				!cptr->hostbls : 0) : 1))
		ClearDNSBL(cptr);
}

static void dnsbl_remove(aDNSBL *bl, int free)
{
	if (bl->prev)
		bl->prev->next = bl->next;
	if (bl->next)
		bl->next->prev = bl->prev;
	if (dnsbls == bl)
		dnsbls = bl->next;

	if (free)
		free_dnsbl(bl);
}

static int query_one_dnsbl(aClient *cptr, aDNSBL *bl, int mask)
{
	Link lin;
	static char hname[HOSTLEN + 1];
	struct hostent *hp;
	u_long ip;
	u_char *ipo = (u_char *) &ip;

	ip = htonl(cptr->ip.s_addr);
	if (mask == DNSBL_IP)
		ircsnprintf(hname, HOSTLEN + 1, "%d.%d.%d.%d.%s", ipo[0],
			ipo[1], ipo[2], ipo[3], bl->zone);
	else if (cptr->hostp)
		ircsnprintf(hname, HOSTLEN + 1, "%s.%s", cptr->hostp->h_name,
			bl->zone);
	else
		/* TODO: Show error here? -rzhe */
		return 0;

	lin.flags = (mask == DNSBL_IP ? ASYNC_DNSBLIP : ASYNC_DNSBLHOST);
	lin.value.cptr = cptr;
	lin.next = (struct SLink *) bl;

	Debug((DEBUG_DNS, "lookup %s", hname));

	if (mask == DNSBL_IP)
		cptr->ipbls++;
	else
		cptr->hostbls++;
	bl->clients++;

	SetDNSBL(cptr);

	if ((hp = gethost_byname(hname, &lin)))
		return check_blacklisted(cptr, bl, hp, lin.flags);
	else
		return 0;
}

static void reject_blacklisted(aClient *cptr)
{
	static char buf[16];

	struct userBan *ban;
	int mask, local = cptr->blflags & BLFLAGS_LISTEDLOCAL;

	strcpy(buf, inetntoa((char *) &cptr->ip));

	if ((ban = make_hostbased_ban("*", buf)))
	{
		mask = (local ? UBAN_LOCAL : UBAN_NETWORK) | UBAN_TEMPORARY;
		ban->flags |= mask | UBAN_DNSBL;

		if (!find_userban_exact(ban, mask))
		{
			sendto_realops_lev(REJ_LEV, "Placing a 1d "
				"blacklisted host %s-ban for *@%s",
				local ? "kI" : "aI", buf);

			ban->reason = (char *) MyMalloc(sizeof(denyreason));
			strcpy(ban->reason, denyreason);
			ban->timeset = NOW;
			ban->duration = 86400;

			add_hostbased_userban(ban);
		}
		else
			userban_free(ban);
	}

	sendto_realops_lev(REJ_LEV, "Rejecting connection from blacklisted "
		"host: %s", buf);

	throttle_force(buf);
	ircstp->is_ref++;

	sendto_one(cptr, ":%s NOTICE * :*** Notice -- %s", me.name,
		denyreason);

	exit_client(cptr, cptr, &me, "Blacklisted");
}

void rehash_dnsbls()
{
	Link *lp;
	aDNSBL *bl, *nextbl;
	Link *fl, *nextfl;
	char *ip, *zone;
	int flags;

	bl = dnsbls;
	while (bl)
	{
		bl->flags &= ~DNSBL_ACTIVE;
		bl = bl->next;
	}

	nextfl = ipfilters;
	while ((fl = nextfl))
	{
		nextfl = fl->next;
		MyFree(fl->value.cp);
		free_link(fl);
	}
	ipfilters = NULL;

	nextfl = hostfilters;
	while ((fl = nextfl))
	{
		nextfl = fl->next;
		MyFree(fl->value.cp);
		free_link(fl);
	}
	hostfilters = NULL;

	hbl_count = hf_count = 0;

	while ((lp = l_lines))
	{
		ip = lp->value.aconf->host;
		flags = lp->value.aconf->port;

		if (!(flags & DNSBL_FILTER))
		{
			zone = lp->value.aconf->name;

			bl = dnsbls;
			while (bl)
			{
				if (!strcasecmp(zone, bl->zone))
				{
					bl->exceptip.s_addr = inet_addr(ip);
					strlcpy(bl->zone, zone, ZONELEN);
					bl->flags = flags | DNSBL_ACTIVE;

					break;
				}

				bl = bl->next;
			}

			if (!bl)
			{
				bl = make_dnsbl();

				bl->exceptip.s_addr = inet_addr(ip);
				strlcpy(bl->zone, zone, ZONELEN);
				bl->flags = flags | DNSBL_ACTIVE;

				bl->clients = 0;

				bl->prev = NULL;
				bl->next = dnsbls;
				if (dnsbls)
					dnsbls->prev = bl;
				dnsbls = bl;
			}
		}
		else if (!BadPtr(ip))
		{
			fl = make_link();
			DupString(fl->value.cp, ip);

			if (!ip[strspn(ip, "1234567890.*?")])
			{
				fl->next = ipfilters;
				ipfilters = fl;
			}
			else
			{
				fl->next = hostfilters;
				hostfilters = fl;
				hf_count++;
			}
		}

		l_lines = lp->next;
		free_link(lp);
	}

	nextbl = dnsbls;
	while ((bl = nextbl))
	{
		nextbl = bl->next;

		if (bl->flags & DNSBL_ACTIVE)
		{
			if (bl->flags & DNSBL_HOST)
				hbl_count++;
		}
		else
			dnsbl_remove(bl, !bl->clients);
	}
}

int query_dnsbls(aClient *cptr, int mask)
{
	aDNSBL *bl, *nextbl = dnsbls;
	Link *fl;
	char *host;
	int rc;

	if (mask == DNSBL_HOST)
	{
		if (!DoingDNSBL(cptr))
			return 0;

		cptr->blflags |= BLFLAGS_GOTHOST;

		if (!cptr->hostp)
		{
			cptr->blflags |= BLFLAGS_SENTRHSS;
			update_blacklisted(cptr);

			return 0;
		}
	}

	if (mask == DNSBL_IP)
	{
		fl = ipfilters;
		host = inetntoa((char *) &cptr->ip);
	}
	else
	{
		fl = hostfilters;
		host = (char *) cptr->hostp->h_name;
	}
	while (fl)
	{
		if (!match(fl->value.cp, host))
		{
			cptr->blflags |= BLFLAGS_EXCEPTED;
			ClearDNSBL(cptr);
			return 0;
		}

		fl = fl->next;
	}

	while ((bl = nextbl))
	{
		nextbl = bl->next;

		if (mask == DNSBL_HOST && !nextbl)
			cptr->blflags |= BLFLAGS_SENTRHSS;

		if ((bl->flags & (DNSBL_ACTIVE | mask)) ==
				(DNSBL_ACTIVE | mask))
		{
			if ((rc = query_one_dnsbl(cptr, bl, mask)))
				return rc;

			if (!DoingDNSBL(cptr))
				return 0;
		}
	}

	update_blacklisted(cptr);

	return 0;
}

int check_blacklisted(aClient *cptr, aDNSBL *bl, struct hostent *hp,
		int flags)
{
	if (hp && DoingDNSBL(cptr))
	{
		if (((struct in_addr *) hp->h_addr)->s_addr ==
				bl->exceptip.s_addr)
		{
			cptr->blflags |= BLFLAGS_EXCEPTED;
			ClearDNSBL(cptr);
		}
		else
			cptr->blflags |= BLFLAGS_LISTED |
				(bl->flags & DNSBL_LOCAL);
	}

	if (DoingDNSBL(cptr))
	{
		switch (flags)
		{
		case ASYNC_DNSBLIP:
			cptr->ipbls--;
			break;
		case ASYNC_DNSBLHOST:
			cptr->hostbls--;
			break;
		default:
			/* TODO: Error message? -rzhe */
			break;
		}

		update_blacklisted(cptr);
	}

	if (bl)
		dnsbl_release(bl);

	if (!DoingDNSBL(cptr) && ((cptr->blflags & BLFLAGS_FOUND) ==
				BLFLAGS_LISTED))
	{
		del_queries((char *) cptr);

		if (DoingAuth(cptr))
		{
			close(cptr->authfd);
			if (cptr->authfd == highest_fd)
				while (!local[highest_fd])
					highest_fd--;
			cptr->authfd = -1;
			ClearAuth(cptr);
		}

		reject_blacklisted(cptr);

		return 1;
	}
	else
		return 0;
}

void dnsbl_release(aDNSBL *bl)
{
	bl->clients--;
	if (!(bl->flags & DNSBL_ACTIVE) && !bl->clients)
		dnsbl_remove(bl, 1);
}
