/************************************************************************
 *   IRC - Internet Relay Chat, src/drone.c
 *   Copyright (C) 2002, DALnet coding team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "throttle.h"
#include "h.h"
#include "userban.h"

#include <dlfcn.h>


#define DRONEMODULE_NAME		"drone.so"
#define DRONEMODULE_INTERFACE_VERSION	0x0202

#define DMS_LOADED			0x0001

#define DCS_WARN			0x0001
#define DCS_WSQUELCH			0x0002
#define DCS_SSQUELCH			0x0004
#define DCS_AIBAN1D			0x0008
#define DCS_REJECT			0x0010

#define DSTYLE_GENERIC			"generic"


static const char denyreason[] = "You match the pattern of a known trojan, please check your system";

static void *dm_handle = NULL;
static int dm_status = 0;

static int (*dm_on_registered_func)(void **pdmdata, const char *n,
	const char *u, const char *h, const char *g, char **dstyle) = NULL;
static void (*dm_version_func)(char **v, char **sv) = NULL;
static void (*dm_cleanup_func)(void **pdmdata) = NULL;


static int load_drone_module(char *errstr, size_t errlen)
{
    void (*dm_ifver_func)(int *v) = NULL;
    char buf[MAXPATHLEN];
    int ver;
    struct Message *mptr;

    if (dm_status & DMS_LOADED)
    {
	snprintf(errstr, errlen, "Drone module is already loaded");
	return 0;
    }

    snprintf(buf, MAXPATHLEN, "%s/%s/%s", dpath, MDIR, DRONEMODULE_NAME);
#ifndef OS_OPENBSD
    if (!(dm_handle = dlopen(buf, RTLD_NOW)))
#else
    if (!(dm_handle = dlopen(buf, DL_LAZY)))
#endif
    {
	snprintf(errstr, errlen, "Error loading " DRONEMODULE_NAME ": %s",
			dlerror());
	return 0;
    }

    if (!(dm_ifver_func = dlsym(dm_handle, "get_interface_version")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			DRONEMODULE_NAME ": %s", dlerror());
	dlclose(dm_handle);
	dm_handle = NULL;
	return 0;
    }

    (*dm_ifver_func)(&ver);
    if (ver != DRONEMODULE_INTERFACE_VERSION)
    {
	snprintf(errstr, errlen, DRONEMODULE_NAME " uses interface version "
			"%d, my interface version is %d", ver,
			DRONEMODULE_INTERFACE_VERSION);
	dlclose(dm_handle);
	dm_handle = NULL;
	return 0;
    }

    if (!(dm_version_func = dlsym(dm_handle, "get_module_version")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			DRONEMODULE_NAME ": %s", dlerror());
	dlclose(dm_handle);
	dm_handle = NULL;
	return 0;
    }

    if (!(dm_on_registered_func = dlsym(dm_handle, "on_registered")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			DRONEMODULE_NAME ": %s", dlerror());
	dlclose(dm_handle);
	dm_handle = NULL;
	return 0;
    }

    if (!(dm_cleanup_func = dlsym(dm_handle, "cleanup")))
    {
	snprintf(errstr, errlen, "Error loading functions in "
			DRONEMODULE_NAME ": %s", dlerror());
	dlclose(dm_handle);
	dm_handle = NULL;
	return 0;
    }

    mptr = msgtab;
    while (mptr->cmd)
    {
	snprintf(buf, sizeof(buf), "on_%s", mptr->cmd);
	mptr->dmfunc = dlsym(dm_handle, buf);

	mptr++;
    }

    dm_status |= DMS_LOADED;

    return 1;
}

static int unload_drone_module(char *errstr, size_t errlen)
{
    struct Message *mptr;
    int i;
    aClient *cptr;

    if (!(dm_status & DMS_LOADED))
    {
	snprintf(errstr, errlen, "Drone module is not loaded");
	return 0;
    }

    mptr = msgtab;
    while (mptr->cmd)
    {
	mptr->dmfunc = NULL;
	mptr++;
    }

    for (i = 0; i <= highest_fd; i++)
	if ((cptr = local[i]) && MyConnect(cptr) && cptr->dmdata)
	    (*dm_cleanup_func)(&cptr->dmdata);

    dlclose(dm_handle);
    dm_handle = NULL;
    dm_status &= ~DMS_LOADED;

    return 1;
}

void drone_init()
{
    char errbuf[256], *ver;

    if (!load_drone_module(errbuf, sizeof(errbuf)))
    {
	fprintf(stderr, "%s\n", errbuf);
	return;
    }

    (*dm_version_func)(&ver, NULL);
    printf("Loaded module " DRONEMODULE_NAME ", version %s\n",
		    ver ? ver : "unspecified");
}

void drone_rehash(int sig)
{
    char errbuf[256], *ver;

    if (sig)
	sendto_ops("Got signal SIGUSR1, reloading drone module");

    if (!unload_drone_module(errbuf, sizeof(errbuf)))
	sendto_realops(errbuf);

    if (load_drone_module(errbuf, sizeof(errbuf)))
    {
	(*dm_version_func)(&ver, NULL);
	sendto_realops("Successfully loaded module " DRONEMODULE_NAME ". "
			"Version: %s", ver ? ver : "unspecified");
    }
    else
	sendto_realops(errbuf);
}

int drone_check(aClient *cptr, int parc, char *parv[], int (*dm_func) ())
{
    char *dstyle = NULL;
    int rc;

    if (!(dm_status & DMS_LOADED))
	return 0;

    rc = dm_func ? (*dm_func)(&cptr->dmdata, parc, parv, &dstyle) :
	(*dm_on_registered_func)(&cptr->dmdata, cptr->name,
	    cptr->user->username, cptr->user->host, cptr->info, &dstyle);

    if (rc)
    {
	char buf[HOSTLEN + USERLEN + HOSTLEN + REALLEN + 8];
	char ipbuf[HOSTIPLEN + 1], *hostip;

	if (cptr->hostip[0] != '\0')
	    hostip = cptr->hostip;
	else
	{
	    strlcpy(ipbuf, inetntoa((char *) &cptr->ip), sizeof(ipbuf));
	    hostip = ipbuf;
	}

	if (cptr->user)
	    snprintf(buf, sizeof(buf), "%s (%s@%s(%s)) [%s]", cptr->name,
		cptr->user->username, cptr->user->host, hostip, cptr->info);
	else
	    snprintf(buf, sizeof(buf), "%s (%s) [%s]", cptr->name, hostip,
		cptr->info);

	if (!cptr->ip.s_addr)
	{
	    sendto_realops("Drone (%s, %.4X) caught at zero-ip client: %s",
		dstyle ? dstyle : DSTYLE_GENERIC, rc, buf);
	    return 0;
	}

	if (rc & DCS_WARN)
	    sendto_realops_lev(REJ_LEV, "Drone warning (%s): %s",
		dstyle ? dstyle : DSTYLE_GENERIC, buf);

	if (rc & DCS_WSQUELCH)
	{
	    sendto_realops_lev(REJ_LEV, "WSquelching drone/spammer (%s): %s",
		dstyle ? dstyle : DSTYLE_GENERIC, buf);

	    SetWSquelch(cptr);
	}

	if (rc & DCS_SSQUELCH)
	{
	    sendto_realops_lev(REJ_LEV, "SSquelching drone/spammer (%s): %s",
		dstyle ? dstyle : DSTYLE_GENERIC, buf);

	    SetSSquelch(cptr);
	}

	if (rc & DCS_AIBAN1D)
	{
	    struct userBan *ban;

	    if ((ban = make_hostbased_ban("*", hostip)))
	    {
		ban->flags |= UBAN_NETWORK | UBAN_TEMPORARY | UBAN_DRONE;

		if (!find_userban_exact(ban, UBAN_NETWORK | UBAN_TEMPORARY))
		{
		    sendto_realops_lev(REJ_LEV, "Placing a 1d drone aI-ban "
			"for *@%s", hostip);

		    ban->reason = (char *) MyMalloc(sizeof(denyreason));
		    strcpy(ban->reason, denyreason);
		    ban->timeset = NOW;
		    ban->duration = 86400;

		    add_hostbased_userban(ban);
		}
		else
		    userban_free(ban);
	    }
	}

	if (rc & DCS_REJECT)
	{
	    sendto_realops_lev(REJ_LEV, "Rejecting %s-style drone: %s",
		dstyle ? dstyle : DSTYLE_GENERIC, buf);

	    throttle_force(hostip);
	    ircstp->is_ref++;
	    ircstp->is_drone++;

	    sendto_one(cptr, ":%s NOTICE %s :*** Notice -- %s", me.name,
		(BadPtr(cptr->name) ? "*" : cptr->name), denyreason);

	    return 1;
	}
    }

    return 0;
}

void drone_cleanup(aClient *cptr)
{
    if (dm_status & DMS_LOADED && MyConnect(cptr) && cptr->dmdata)
	(*dm_cleanup_func)(&cptr->dmdata);
}

char *drone_version(char *buf, int buflen)
{
    char *ver = NULL;

    if (!(dm_status & DMS_LOADED))
    {
	ircsnprintf(buf, buflen, "No drone module is currently installed.");
	return buf;
    }

    (*dm_version_func)(&ver, NULL);
    ircsnprintf(buf, buflen, "%s", ver ? ver : "unspecified");
    return buf;
}

char *drone_short_version(char *buf, int buflen)
{
    char *ver = NULL;

    if (!(dm_status & DMS_LOADED))
    {
	*buf = '\0';
	return NULL;
    }

    (*dm_version_func)(NULL, &ver);
    ircsnprintf(buf, buflen, "%s", ver ? ver : "unspecified");
    return buf;
}
